from typing import TypeAlias

from d3m.core import DomainName

ObjectsCategory: TypeAlias = str
ObjectName: TypeAlias = str
CategoryCollection: TypeAlias = dict[DomainName, dict[ObjectName, type]]


class SingletonMetaclass(type):
    _instance: "SingletonMetaclass"

    def __call__(cls, *args, **kwargs):
        if not hasattr(cls, "_instance"):
            cls._instance = super().__call__(*args, **kwargs)
        return cls._instance


class DomainObjectsCollection(metaclass=SingletonMetaclass):
    def __init__(self):
        self._collections: dict[ObjectsCategory, CategoryCollection] = {}
        self._registered_classes: set[int] = set()

    def register(
        self,
        category: ObjectsCategory,
        domain: DomainName | str,
        name: ObjectName,
        klass: type,
    ):
        assert isinstance(
            klass, type
        ), "Allows registration of only classes, not instance"
        domain = DomainName(domain)
        category = category.lower()
        klass_id = id(klass)
        registered_class = self._get_class(category, domain, name)
        if klass_id in self._registered_classes and registered_class is not klass:
            raise RuntimeError(
                f"{klass!r} already registered in collection with other name"
            )
        if registered_class is not None and registered_class is not klass:
            raise RuntimeError(
                f'Other {category} class for "{domain}" domain with name "{name}" already registered'
            )
        self._get_domain_collection(category, domain)[name] = klass
        self._registered_classes.add(klass_id)

    def get_domain_objects(
        self, category: ObjectsCategory, domain: DomainName | str
    ) -> tuple[type, ...]:
        category = category.lower()
        domain = DomainName(domain)
        if (
            category not in self._collections
            or domain not in self._collections[category]
        ):
            return ()
        return tuple(self._collections[category][domain].values())  # type: ignore

    def get_domain_object(
        self, category: ObjectsCategory, domain: DomainName | str, name: ObjectName
    ):
        category = category.lower()
        domain = DomainName(domain)
        klass = self._get_class(category, domain, name)
        if klass is None:
            raise RuntimeError(
                f'{category} class for "{domain}" domain with name "{name}" not registered'
            )
        return klass

    def _get_domain_collection(
        self, category: ObjectsCategory, domain: DomainName
    ) -> dict[ObjectName, type]:
        return self._collections.setdefault(category, {}).setdefault(domain, {})

    def _get_class(
        self, category: ObjectsCategory, domain: DomainName, name: ObjectName
    ) -> type | None:
        return self._collections.get(category, {}).get(domain, {}).get(name, None)



from uuid import UUID, uuid4

from pydantic import BaseModel
from typing import Generic, Callable, Iterable, Any, NewType
from pydantic._internal._model_construction import ModelMetaclass  # noqa

from d3m.core import (
    MessageName,
    IEntity,
    IRootEntity,
    AbstractEvent,
    DomainName,
)
from d3m.core.abstractions import Version, IRootEntityMeta, _ReferenceType
from . import get_event_class
from .bases import get_domain_name


class _EntityMeta(ModelMetaclass):
    _reference_factory: Callable[[], Any] | None = None
    _reference_type: type | None = None

    def __init__(cls, name, bases, namespace, *, reference_factory=None, **kwargs):
        super().__init__(name, bases, namespace, **kwargs)
        generic = cls.__pydantic_generic_metadata__.get("args", ())  # type: ignore[attr-defined]
        if not generic:
            cls._reference_type = cls._reference_type or UUID
        elif type(generic[0]) is NewType:
            cls._reference_type = generic[0].__supertype__  # type: ignore[assignment]
        else:
            cls._reference_type = generic[0]

        if reference_factory is not None:
            cls._reference_factory = staticmethod(reference_factory)
        elif cls._reference_type is UUID:
            cls._reference_factory = staticmethod(cls._reference_factory or uuid4)

    def __call__(cls, *, __reference__=None, **date):
        obj = super().__call__(**date)
        if __reference__ is None and obj._reference_factory is not None:
            __reference__ = obj._reference_factory()
        if not isinstance(__reference__, obj._reference_type):
            raise TypeError(f"Invalid reference type. Expected {obj._reference_type!r}")
        obj.__pydantic_private__["_Entity__reference"] = __reference__
        return obj


class Entity(IEntity, BaseModel, Generic[_ReferenceType], metaclass=_EntityMeta):
    """
    Class Entity

    A base class representing an entity.

    Attributes:
        __reference__ (_ReferenceType): The reference of the entity.

    Examples:
        **Create an entity class:**
        >>> class Person(Entity):
        ...     name: str
        ...     surname: str
        ...
        >>> person = Person(name="John", surname="Doe")
        >>> assert isinstance(person.__reference__, UUID)

        by default reference type is `UUID` and default factory is `uuid.uuid4()`

        **Set custom reference value:**
        >>> import uuid
        >>> person = Person(name="John", surname="Doe", __reference__=uuid.UUID(int=1))
        >>> assert person.__reference__ == uuid.UUID(int=1)

        **Set custom reference type:**
        >>> PersonId = int
        >>> class Person(Entity[PersonId]):
        ...     name: str
        ...     surname: str
        >>> # when set custom reference type, the default factory is not set
        >>> person = Person(name="John", surname="Doe", __reference__=1)
        >>> assert person.__reference__ == 1

        **Set default factory:**
        >>> import random
        >>> class Person(Entity[PersonId], default_factory=random.randrange(1, 1000000)):
        ...     name: str
        ...     surname: str
        >>> person = Person(name="John", surname="Doe")
        >>> assert isinstance(person.__reference__, int)
        >>> assert 1 <= person.__reference__ < 1000000

    """

    __reference: _ReferenceType

    def __init_subclass__(cls, *, reference_factory=None, **kwargs):
        super().__init_subclass__(**kwargs)

    @property
    def __reference__(self) -> _ReferenceType:
        """
        Returns the reference of the entity.

        Returns:
            _ReferenceType: The reference of the entity.
        """
        return self.__reference

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and self.__reference__ == other.__reference__
        )

    def __hash__(self):
        return hash(self.__reference__)


class _RootEntityMeta(_EntityMeta, IRootEntityMeta, ModelMetaclass):
    def __init__(cls, name, bases, namespace, *, domain: str | None = None, **kwargs):
        super().__init__(name, bases, namespace, **kwargs)
        domain = get_domain_name(cls, bases, domain)
        if domain is not None:
            cls.__domain_name = domain
        elif domain is None and cls.__module__ != __name__:
            raise ValueError(
                f"required set domain name for root entity '{cls.__module__}.{cls.__name__}'"
            )

    def __call__(cls, __version__=Version(1), **data):
        obj = super().__call__(**data)
        obj.__pydantic_private__["_RootEntity__version"] = __version__
        return obj

    @property
    def __domain_name__(cls) -> DomainName:
        return cls.__domain_name


class RootEntity(
    Entity,
    IRootEntity,
    Generic[_ReferenceType],
    metaclass=_RootEntityMeta,
):
    """
    RootEntity

    A root entity is a base class for domain entities that serves as the root of an aggregate.
     It provides common functionality for managing events and versioning.

    Attributes:
        __reference__ (_ReferenceType): the reference of the root entity.
        __domain_name__ (DomainName): the domain name associated with the root entity
        __version__ (Version): The current version of the root entity.

    Methods:
        create_event: Create a domain event for this entity.
        collect_events: Collect all the events that have been created for this entity.

    Examples:
        **Set domain of root entity:**
        >>> class Person(RootEntity, domain='person'):
        ...     name: str
        ...     surname: str
        >>> assert Person.__domain_name__ == DomainName('person')
        >>> person = Person(name="John", surname="Doe")
        >>> assert person.__domain_name__ == DomainName('person')
        >>> assert isinstance(person.__reference__, UUID)
        >>> assert person.__version__ == 1

        **Set custom version of root entity:**
        >>> person = Person(name="John", surname="Doe", __version__=2)
        >>> assert person.__version__ == 2

        **Create and collect domain events for this entity:**
        >>> events = list(person.collect_events())
        >>> assert len(events) == 0
        >>> person.create_event('PersonCreated',
        ...                     reference=person.__reference__,
        ...                     name=person.name,
        ...                     surname=person.surname)
        >>> events = list(person.collect_events())
        >>> assert len(events) == 1
        >>> event = events[0]
        >>> assert isinstance(event, AbstractEvent)
        >>> assert event.__domain_name__ == DomainName('person')
        >>> assert event.__message_name__ == MessageName('PersonCreated')
        >>> events = list(person.collect_events())
        >>> assert len(events) == 0

    """

    __version: Version = Version(1)

    def __init_subclass__(cls, *, domain: str | None = None, **kwargs):
        super().__init_subclass__()

    def __init__(self, **data):
        super().__init__(**data)
        self._events: set[AbstractEvent] = set()

    @property
    def __domain_name__(self) -> DomainName:
        """
        Get the domain name associated with the current class.

        Returns:
            DomainName: The domain name associated with the class.

        """
        return self.__class__.__domain_name__  # type: ignore

    @property
    def __version__(self) -> Version:
        """
        Get the current version of the root entity.

        Returns:
             Version: The current version of the root entity.
        """
        return self.__version

    def create_event(self, __name: MessageName | str, /, **payload):
        """
        Create a domain event for this root entity.

        Attributes:
            __name (MessageName|str): The name of the event. Event class required be declared for domain of the root entity.
            **payload: Additional keyword arguments to be passed as payload data for creating the event.

        """
        event_class = get_event_class(self.__domain_name__, __name)
        event = event_class.load(payload)
        self._events.add(event)

    def collect_events(self) -> Iterable[AbstractEvent]:
        """
        Collects and returns events from this root entity.

        Returns:
            An iterable of `AbstractEvent` objects.
        """
        while self._events:
            yield self._events.pop()


def increment_version(entity: RootEntity):
    """
    Increments the version of the root entity

    Attributes:
        entity (RootEntity): The root entity object whose version needs to be incremented.

    """
    if isinstance(entity.__pydantic_private__, dict):
        version = entity.__pydantic_private__.get("_RootEntity__version")
        if version:
            version += 1
        entity.__pydantic_private__["_RootEntity__version"] = version



from .event import DomainEvent, get_event_class
from .command import DomainCommand, get_command_class
from .exceptions import (
    DomainError,
    get_error_class,
    get_or_create_error_class,
    get_or_create_base_error_class,
)
from .entities import Entity, RootEntity

__all__ = [
    "Entity",
    "RootEntity",
    "DomainEvent",
    "get_event_class",
    "DomainCommand",
    "get_command_class",
    "DomainError",
    "get_error_class",
    "get_or_create_error_class",
    "get_or_create_base_error_class",
]



import inspect
from contextlib import suppress
from string import Formatter
from typing import Mapping, Any

from d3m.core import DomainName
from d3m.core.types import FrozenJsonDict
from .bases import get_domain_name
from .collection import DomainObjectsCollection


def _get_template_parameters(template: str) -> tuple[str, ...]:
    result = []
    for _, name, _, _ in Formatter().parse(template):
        if name is None:
            continue
        if not name.isidentifier():
            raise ValueError(f"invalid template parameter name {name!r}")
        result.append(name)
    return tuple(result)


class _DomainErrorMeta(type):
    def __init__(cls, name, bases, namespace, domain: DomainName | str | None = None):
        super().__init__(name, bases, namespace, domain=domain)
        if cls.__module__ == __name__ and cls.__name__ == "DomainError":
            return
        domain = get_domain_name(cls, bases, domain)
        if domain is not None:
            cls.__domain_name = domain
        else:
            raise ValueError(
                f"required set domain name for error '{cls.__module__}.{cls.__name__}'"
            )
        if (
            DomainError in bases
            and (base_class := _get_base_domain_error_class(cls.__domain_name__))
            is not None
        ):
            raise RuntimeError(
                f"base error class for domain '{cls.__domain_name__}' "
                f"already registered: {base_class.__module__}.{base_class.__qualname__}"
            )

        if hasattr(cls, "__template__"):
            cls.__init_parameters = _get_template_parameters(cls.__template__)

        DomainObjectsCollection().register(
            "error", cls.__domain_name__, cls.__name__, cls
        )

    @property
    def __domain_name__(cls) -> DomainName:
        return cls.__domain_name

    def validate_init_parameters(cls, parameters: dict):
        params = [param for param in cls.__init_parameters if param not in parameters]
        params_count = len(params)
        error_message = None
        if params_count == 1:
            error_message = (
                f"{cls.__module__}.{cls.__qualname__}() missing {params_count} "
                f"required keyword-only argument: {params[0]!r}"
            )
        elif params_count > 1:
            last_param = params[-1]
            other_params = ", ".join(map(repr, params[:-1]))
            error_message = (
                f"{cls.__module__}.{cls.__qualname__}() missing {params_count} "
                f"required keyword-only arguments: {other_params} and {last_param!r}"
            )

        if error_message is not None:
            raise TypeError(error_message)


class DomainError(Exception, metaclass=_DomainErrorMeta):
    """
    Custom Exception class for domain-specific errors

    It allows for the definition of specific error messages and
    supports dynamic parameter substitution in
    the error message template.

    Attributes:
        __template__ (str):
        The template string used for formatting the error message.
        All attributes in template required for init error.
        __domain_name__ (DomainName): readonly attribute retrieves the domain name associated with the current class.
        __payload__ (Mapping[str, Any]): readonly attribute retrieves the payload (additional parameters) associated with the error.

    Examples:
        **Creating a custom DomainError subclass**
        >>> class MyDomainError(DomainError, domain='my-domain'):
        ...     __template__ = "An error occurred in domain '{domain}'."
        >>> assert MyDomainError.__domain_name__ == DomainName('my-domain')

        **Generating error message use template**
        >>> MyDomainError(domain='my-domain', foo='bar')
        MyDomainError("An error occurred in domain 'test'.")

        **Get initialization error attribures**
        >>> MyDomainError(domain='my-domain', foo='bar').__payload__
        {'domain': 'my-domain', 'foo': 'bar'}

        **Fail init error without template attributes**
        >>> MyDomainError(foo='bar')
        TypeError: MyDomainError() missing 1 required keyword-only argument: 'domain'

        **Init with custom message, ignore template**
        >>> error = MyDomainError('custom message')
        >>> error
        MyDomainError('custom message')
        >>> error.__payload__
        {}
    """

    __template__: str

    def __init_subclass__(cls, domain: DomainName | str | None = None, **kwargs):
        super().__init_subclass__(**kwargs)

    def __init__(self, __message: str | None = None, /, **kwargs):
        if __message is None:
            self.__class__.validate_init_parameters(kwargs)
            message = self.__template__.format_map(kwargs)
        else:
            message = __message
        super().__init__(message)
        self._payload = FrozenJsonDict(kwargs)

    @property
    def __domain_name__(self) -> DomainName:
        return self.__class__.__domain_name__  # type: ignore

    @property
    def __payload__(self) -> Mapping[str, Any]:
        return self._payload


def get_error_class(domain: str | DomainName, name: str) -> type[DomainError]:
    """
    Return registered error class by domain and name

    Attributes:
        domain (DomainName | str): The domain of the error.
        name (str): The name of the error.

    Returns:
        The class of the error.

    Examples:
        >>> from d3m.domain import DomainError
        >>> class PersonNotFound(DomainError, domain='person')
        ...     __template__: str = 'Person {reference} not found'
        >>> error_class = get_error_class('person', 'PersonNotFound')
        >>> assert error_class is PersonNotFound
    """
    return DomainObjectsCollection().get_domain_object("error", domain, name)


def _get_base_domain_error_class(domain: str | DomainName) -> type[DomainError] | None:
    classes = DomainObjectsCollection().get_domain_objects("error", domain)
    return next((klass for klass in classes if DomainError in klass.__bases__), None)  # type: ignore


def _create_error_class(
    domain: str | DomainName,
    name: str,
    module: str,
    bases=(DomainError,),
    template: str | None = None,
) -> type[DomainError]:
    namespace = {"__module__": module}
    if template is not None:
        namespace["__template__"] = template
    return type(name, bases, namespace, domain=domain)  # type: ignore


def get_or_create_base_error_class(
    domain: str | DomainName, *, template: str | None = None
) -> type[DomainError]:
    """
    Get or create a base error class associated with the given domain

    Attributes:
        domain (str | DomainName): The domain of the base error class.
        template (str | None): The template for the base error class. Defaults to None.
    Returns:
        The base error class associated with the given domain, or a newly created base error class if one does not exist.

    Examples:
        **Get existing base error class**
        >>> class BasePersonDomainError(DomainError, domain='person'):
        ...     pass
        >>> class PersonNotFound(BasePersonDomainError):
        ...     __template__ = 'Person {reference} not found'
        >>> error_class = get_or_create_base_error_class('person')
        >>> assert error_class is BasePersonDomainError

        **Create new base error class**
        >>> error_class = get_or_create_base_error_class('person')
        >>> assert error_class.__module__ is __name__
        >>> error_class.__name__
        '__BaseError__'
        >>> error_class.__domain_name__
        DomainName('person')
    """
    mod = inspect.getmodule(inspect.stack()[1].frame)
    module = mod.__name__ if mod else __name__
    return _get_or_create_base_error_class(domain, module=module, template=template)


def _get_or_create_base_error_class(
    domain: str | DomainName,
    *,
    template: str | None = None,
    module: str,
) -> type[DomainError]:
    base = _get_base_domain_error_class(domain)
    if base is not None:
        return base
    return _create_error_class(domain, "__BaseError__", module, template=template)


def get_or_create_error_class(
    domain: str | DomainName, name: str, template: str | None = None
) -> type[DomainError]:
    """
    Get or create a error class associated with the given domain.
    If base error class associated with the given domain does not exist, then create it.

    Attributes:
        domain (str | DomainName): The domain of the error class.
        template (str | None): The template for the error class. Defaults to None.

    Returns:
        A type object representing the error class for the given domain and name.

    Examples:
        **Get existing error class**
        >>> class BasePersonDomainError(DomainError, domain='person'):
        ...     pass
        >>> class PersonNotFound(BasePersonDomainError):
        ...     __template__ = 'Person {reference} not found'
        >>> error_class = get_or_create_error_class(
        ...     domain='person',
        ...     name='PersonNotFound',
        ... )
        >>> assert error_class is PersonNotFound

        **Create new error class**
        >>> class BasePersonDomainError(DomainError, domain='person'):
        ...     pass
        >>> error_class = get_or_create_error_class(
        ...     domain='person',
        ...     name='PersonNotFound',
        ...     template='Person {reference} not found'
        ... )
        >>> assert error_class.__name__ == 'PersonNotFound'
        >>> assert error_class.__domain_name__ = 'person'
        >>> assert isinstance(error_class, BasePersonDomainError)
    """
    with suppress(RuntimeError):
        return get_error_class(domain, name)

    mod = inspect.getmodule(inspect.stack()[1].frame)
    module = mod.__name__ if mod else __name__
    base = _get_or_create_base_error_class(domain, template=None, module=module)
    return _create_error_class(domain, name, module, (base,), template)



import abc
import datetime as dt
from contextlib import suppress
from typing import Mapping, TypeVar, no_type_check, Type
from uuid import UUID, uuid4

from pydantic import BaseModel, PrivateAttr, ConfigDict
from pydantic._internal._model_construction import ModelMetaclass  # noqa

from d3m.core import IMessageMeta, MessageName, DomainName, IMessage
from .collection import DomainObjectsCollection


def get_domain_name(
    cls,
    bases: tuple[type[IMessageMeta]],
    domain: DomainName | str | None = None,
) -> DomainName | None:
    domains: set[DomainName] = set()  # type: ignore
    for base in bases:
        with suppress(AttributeError):
            if base.__domain_name__ is not None:
                domains.add(base.__domain_name__)  # type: ignore

    if len(domains) > 1:
        raise RuntimeError("Not allowed multiple inheritance domain")

    if len(domains) == 1 and domain is not None and domain != cls.__domain_name__:
        raise RuntimeError(
            f"not allowed replace domain name in child class: {cls.__module__}.{cls.__name__}"
        )

    if len(domains) == 0 and domain is not None:
        return DomainName(domain)

    if len(domains) == 1 and (domain is None or domain == cls.__domain_name__):
        return cls.__domain_name__  # type: ignore

    return None


_T = TypeVar("_T", bound="BaseDomainMessage")


class BaseDomainMessageMeta(IMessageMeta, ModelMetaclass, abc.ABCMeta):
    __domain_name: DomainName

    @no_type_check
    def __new__(mcs, name, bases, namespace, domain: str | None = None):
        cls = super().__new__(mcs, name, bases, namespace)
        domain = get_domain_name(cls, bases, domain)
        if domain is not None:
            cls.__domain_name = domain
        cls.model_config["frozen"] = True
        return cls

    def __init__(cls, name, bases, namespace, *, domain: str | None = None):
        super().__init__(name, bases, namespace, domain=domain)

        cls.__message_name = MessageName(name)

        with suppress(AttributeError):
            DomainObjectsCollection().register(
                category=cls.__type__,
                domain=cls.__domain_name__,
                name=cls.__message_name__,
                klass=cls,
            )

    @property
    def __domain_name__(cls) -> DomainName:
        return cls.__domain_name

    @property
    def __message_name__(cls) -> MessageName:
        return cls.__message_name

    @property
    def __payload_model__(cls) -> Type[BaseModel]:
        return cls  # type: ignore[return-value]

    @no_type_check
    def load(
        cls: type[_T],
        payload: Mapping | str | bytes,
        reference: UUID | None = None,
        timestamp: dt.datetime | None = None,
    ) -> _T:
        obj = cls.model_validate(payload)
        obj.__pydantic_private__["_BaseDomainMessage__reference"] = (
            reference or obj.__reference__
        )
        obj.__pydantic_private__["_BaseDomainMessage__timestamp"] = (
            timestamp or obj.__timestamp__
        )
        return obj


class BaseDomainMessage(BaseModel, IMessage, abc.ABC, metaclass=BaseDomainMessageMeta):
    """
    Class representing a base domain message.

    Attributes:
        __domain_name__ (DomainName): The domain name associated with the current class.
        __message_name__ (MessageName): The message name associated with the current class.
        __payload__ (dict): The payload of the message.
        __reference__ (UUID): The unique identifier for the message.
        __timestamp__ (datetime.datetime): The timestamp of when the message was created.

    Methods:
        to_dict: Returns a dictionary representation of the message.
        to_json: Returns a JSON string representation of the message.

    """

    __reference: UUID = PrivateAttr(default_factory=uuid4)
    __timestamp: dt.datetime = PrivateAttr(
        default_factory=lambda: dt.datetime.now(dt.timezone.utc)
    )

    model_config = ConfigDict(frozen=True)

    @property
    def __domain_name__(self) -> DomainName:
        return self.__class__.__domain_name__  # type: ignore

    @property
    def __message_name__(self) -> MessageName:
        return self.__class__.__message_name__  # type: ignore

    @property
    def __reference__(self) -> UUID:
        return self.__reference

    @property
    def __timestamp__(self) -> dt.datetime:
        return self.__timestamp

    @property
    def __payload__(self) -> dict:
        return self.model_dump()

    def to_dict(self) -> dict:
        return self.model_dump(mode="json")

    def to_json(self) -> str:
        return self.model_dump_json()



from d3m.core import (
    MessageType,
    AbstractCommand,
)

from pydantic._internal._model_construction import ModelMetaclass  # noqa

from d3m.core.abstractions import AbstractCommandMeta
from .collection import DomainObjectsCollection
from .bases import BaseDomainMessageMeta, BaseDomainMessage


class _DomainCommandMeta(BaseDomainMessageMeta, AbstractCommandMeta):
    def __init__(cls, name, bases, namespace, *, domain: str | None = None):
        super().__init__(name, bases, namespace, domain=domain)
        if domain is None and cls.__module__ != __name__:
            try:
                _ = cls.__domain_name__
            except AttributeError:
                raise ValueError(
                    f"required set domain name for command '{cls.__module__}.{cls.__name__}'"
                )


class DomainCommand(BaseDomainMessage, AbstractCommand, metaclass=_DomainCommandMeta):
    """
    Class representing a base domain command.

    Attributes:
        __domain_name__ (DomainName): The domain name associated with the current command.
        __message_name__ (MessageName): The message name associated with the current command.
        __payload__ (dict): The payload of the command.
        __payload_model__ (Type[BaseModel]): The payload model associated with the current command.
        __type__ (MessageType): Always return `MessageType.COMMAND`
        __reference__ (UUID): The unique identifier for the command.
        __timestamp__ (datetime.datetime): The timestamp of when the command was created.

    Methods:
        to_dict: Returns a dictionary representation payload of the command.
        to_json: Returns a JSON string representation payload of the command.
        load: class method returns new instance of the domain command

    Examples:
        **Create command class:**
        >>> from d3m.domain import DomainCommand
        >>> class CreatePerson(DomainCommand, domain='person'):
        ...     name: str
        ...     surname: str
        >>> assert CreatePerson.__domain_name__ == DomainName('person')
        >>> assert CreatePerson.__message_name__ == MessageName('CreatePerson')
        >>> assert CreatePerson.__type__ == MessageType.COMMAND

        **Create command instance**
        >>> command = CreatePerson(
        ...               name='John',
        ...               surname='Black'
        ...           )
        >>> assert command.__domain_name__ == DomainName('person')
        >>> assert command.__message_name__ == MessageName('CreatePerson')
        >>> assert isinstance(command.__reference__, UUID)
        >>> assert isinstance(command.__timestamp__, datetime)
        >>> assert command.__payload__ == dict(name='John', surname='Black')
        >>> assert command.name == 'John'
        >>> assert command.surname =='Black'

        **Load command from dict**
        >>> payload = dict(name='John', surname='Black')
        >>> reference = uuid4()
        >>> timestamp = datetime.now(timezone.utc)
        >>> command = CreatePerson.load(payload, reference=reference, timestamp=timestamp)
        >>> assert command.__domain_name__ == DomainName('person')
        >>> assert command.__message_name__ == MessageName('CreatePerson')
        >>> assert command.__reference__ == reference
        >>> assert command.__timestamp__ == timestamp
        >>> assert command.__payload__ == payload

        **Load command from json-string**
        >>> payload = '{"name":"John","surname":"Black"}'
        >>> command = CreatePerson.load(payload, reference=uuid4(), timestamp=datetime.now())
        >>> assert command.__domain_name__ == DomainName('person')
        >>> assert command.__message_name__ == MessageName('CreatePerson')
        >>> assert isinstance(command.__reference__, UUID)
        >>> assert isinstance(command.__timestamp__, datetime)
        >>> assert command.name == 'John'
        >>> assert command.surname =='Black'

        **Command's payload to json serializeble dict**
        >>> assert command.to_dict() == {'name': 'John', 'surname': 'Black'}

        **Command's payload to json**
        >>> assert command.to_dict() == '{"name":"John","surname":"Black"}'
    """


def get_command_class(domain: str, name: str) -> AbstractCommandMeta:
    """
    Return registered command class by domain and name

    Attributes:
        domain (DomainName | str): The domain of the event.
        name (MessageName | str): The name of the event.

    Returns:
        The class of the command.

    Examples:
        >>> from d3m.domain import DomainCommand
        >>> class CreatePerson(DomainCommand, domain='person')
        ...     name: str
        ...     surname: str
        >>> command_class = get_command_class('person', 'CreatePerson')
        >>> assert command_class is CreatePerson
    """
    return DomainObjectsCollection().get_domain_object(
        MessageType.COMMAND, domain, name
    )



from typing import TypeVar

from d3m.core import (
    MessageType,
    AbstractEvent,
    DomainName,
)
from pydantic._internal._model_construction import ModelMetaclass  # noqa

from d3m.core.abstractions import AbstractEventMeta
from d3m.domain.collection import DomainObjectsCollection
from d3m.domain.bases import BaseDomainMessageMeta, BaseDomainMessage

_T = TypeVar("_T", bound="DomainEvent")


class _DomainEventMeta(BaseDomainMessageMeta, AbstractEventMeta):
    def __init__(cls, name, bases, namespace, *, domain: str | None = None):
        super().__init__(name, bases, namespace, domain=domain)
        if domain is None and cls.__module__ != __name__:
            try:
                _ = cls.__domain_name__
            except AttributeError:
                raise ValueError(
                    f"required set domain name for event '{cls.__module__}.{cls.__name__}'"
                )


class DomainEvent(BaseDomainMessage, AbstractEvent, metaclass=_DomainEventMeta):
    """
    Class representing a base domain event.

    Attributes:
        __domain_name__ (DomainName): The domain name associated with the current event.
        __message_name__ (MessageName): The message name associated with the current event.
        __payload__ (dict): The payload of the event.
        __type__ (MessageType): Always return `MessageType.EVENT`
        __reference__ (UUID): The unique identifier for the event.
        __timestamp__ (datetime.datetime): The timestamp of when the event was created.

    Methods:
        to_dict: Returns a dictionary representation payload of the event.
        to_json: Returns a JSON string representation payload of the event.
        load: class method returns new instance of the domain event

    Examples:
        **Create event class:**
        >>> from d3m.domain import DomainEvent
        >>> class PersonCreated(DomainEvent, domain='person'):
        ...     reference: UUID
        ...     name: str
        ...     surname: str
        >>> assert PersonCreated.__domain_name__ == DomainName('person')
        >>> assert PersonCreated.__message_name__ == MessageName('PersonCreated')
        >>> assert PersonCreated.__type__ == MessageType.EVENT

        **Create event instance**
        >>> event = PersonCreated(
        ...             reference=person.__reference__
        ...             name=person.name,
        ...             surname=person.surname
        ...         )
        >>> assert event.__domain_name__ == DomainName('person')
        >>> assert event.__message_name__ == MessageName('PersonCreated')
        >>> assert isinstance(event.__reference__, UUID)
        >>> assert isinstance(event.__timestamp__, datetime)
        >>> assert event.__payload__ == dict(reference=person.__reference__, name=person.name, surname=person.surname)
        >>> assert event.reference == person.__reference__
        >>> assert event.name == person.name
        >>> assert event.surname == person.surname

        **Load event from dict**
        >>> payload = dict(
        ...               reference=person.__reference__
        ...               name=person.name,
        ...               surname=person.surname
        ...           )
        >>> event = PersonCreated.load(payload, reference=uuid4(), timestamp=datetime.now())
        >>> assert event.__domain_name__ == DomainName('person')
        >>> assert event.__message_name__ == MessageName('PersonCreated')
        >>> assert isinstance(event.__reference__, UUID)
        >>> assert isinstance(event.__timestamp__, datetime)
        >>> assert event.__payload__ == payload

        **Load event from json-string**
        >>> payload = '{"reference":"00000000-0000-0000-0000-000000000001","name":"John","surname":"Black"}'
        >>> event = PersonCreated.load(payload, reference=uuid4(), timestamp=datetime.now())
        >>> assert event.__domain_name__ == DomainName('person')
        >>> assert event.__message_name__ == MessageName('PersonCreated')
        >>> assert isinstance(event.__reference__, UUID)
        >>> assert isinstance(event.__timestamp__, datetime)
        >>> assert event.name == 'John'
        >>> assert event.surname =='Black'
        >>> assert event.reference == UUID(int=1)

        **Event's payload to json serializeble dict**
        >>> assert event.to_dict() == {'reference': '00000000-0000-0000-0000-000000000001', 'name': 'John', 'surname': 'Black'}

        **Event's payload to json**
        >>> assert event.to_dict() == '{"reference":"00000000-0000-0000-0000-000000000001","name":"John","surname":"Black"}'
    """


def get_event_class(
    domain: DomainName | str, name: DomainName | str
) -> AbstractEventMeta:
    """
    Return registered event class by domain and name

    Attributes:
        domain (DomainName | str): The domain of the event.
        name (MessageName | str): The name of the event.

    Returns:
        The class of the event.


    Examples:
        >>> from d3m.domain import DomainEvent
        >>> class PersonCreated(DomainEvent, domain='person')
        ...     name: str
        ...     surname: str
        >>> event_class = get_event_class('person', 'PersonCreated')
        >>> assert event_class is PersonCreated
    """
    return DomainObjectsCollection().get_domain_object(MessageType.EVENT, domain, name)
